@@ -173,6 +173,7 @@ test_that("URI-decoding with hive partitioning", {
     fs, selector, NULL, fmt,
     partitioning = partitioning
   )
+  schm <- factory$Inspect()
   ds <- factory$Finish(schm)
   expect_scan_result(ds, schm)
 
@@ -590,11 +591,44 @@ test_that("UnionDataset can merge schemas", {
   ds <- open_dataset(list(ds1, ds2), unify_schemas = FALSE)
   expected <- as_tibble(sub_df1) %>%
     union_all(sub_df2 %>% as_tibble() %>% select(x))
-  actual <- ds %>% collect() %>% arrange(x)
+  actual <- ds %>%
+    collect() %>%
+    arrange(x)
   expect_equal(colnames(actual), c("x", "y"))
   expect_equal(actual, expected)
 })
 
+test_that("UnionDataset handles InMemoryDatasets", {
+  sub_df1 <- Table$create(
+    x = Array$create(c(1, 2, 3)),
+    y = Array$create(c("a", "b", "c"))
+  )
+  sub_df2 <- Table$create(
+    x = Array$create(c(4, 5)),
+    z = Array$create(c("d", "e"))
+  )
+
+  ds1 <- InMemoryDataset$create(sub_df1)
+  ds2 <- InMemoryDataset$create(sub_df2)
+  ds <- c(ds1, ds2)
+  actual <- ds %>%
+    arrange(x) %>%
+    compute()
+  expected <- concat_tables(sub_df1, sub_df2)
+  expect_equal(actual, expected)
+})
+
+test_that("scalar aggregates with many batches (ARROW-16904)", {
+  tf <- tempfile()
+  write_parquet(data.frame(x = 1:100), tf, chunk_size = 20)
+
+  ds <- open_dataset(tf)
+  replicate(100, ds %>% summarize(min(x)) %>% pull())
+
+  expect_true(all(replicate(100, ds %>% summarize(min(x)) %>% pull()) == 1))
+  expect_true(all(replicate(100, ds %>% summarize(max(x)) %>% pull()) == 100))
+})
+
 test_that("map_batches", {
   ds <- open_dataset(dataset_dir, partitioning = "part")
 
@@ -604,17 +638,18 @@ test_that("map_batches", {
       filter(int > 5) %>%
       select(int, lgl) %>%
       map_batches(~ summarize(., min_int = min(int))) %>%
-      arrange(min_int),
+      arrange(min_int) %>%
+      collect(),
     tibble(min_int = c(6L, 101L))
   )
 
-  # $num_rows returns integer vector
+  # $num_rows returns integer vector, so we need to wrap it in a RecordBatch
   expect_equal(
     ds %>%
       filter(int > 5) %>%
       select(int, lgl) %>%
-      map_batches(~ .$num_rows, .data.frame = FALSE) %>%
-      unlist() %>% # Returns list because .data.frame is FALSE
+      map_batches(~ record_batch(nrows = .$num_rows)) %>%
+      pull(nrows) %>%
       sort(),
     c(5, 10)
   )
@@ -623,19 +658,121 @@ test_that("map_batches", {
   expect_equal(
     ds %>%
       map_batches(~ count(., part)) %>%
-      arrange(part),
+      arrange(part) %>%
+      collect(),
     tibble(part = c(1, 2), n = c(10, 10))
   )
 
-  # $Take returns RecordBatch, which gets binded into a tibble
+  # $Take returns RecordBatch
   expect_equal(
     ds %>%
       filter(int > 5) %>%
       select(int, lgl) %>%
       map_batches(~ .$Take(0)) %>%
-      arrange(int),
+      arrange(int) %>%
+      collect(),
     tibble(int = c(6, 101), lgl = c(TRUE, TRUE))
   )
+
+  # Do things in R and put back into Arrow
+  expect_equal(
+    ds %>%
+      filter(int < 5) %>%
+      select(int) %>%
+      map_batches(
+        # as_mapper() can't handle %>%?
+        ~ mutate(as.data.frame(.), lets = letters[int])
+      ) %>%
+      arrange(int) %>%
+      collect(),
+    tibble(int = 1:4, lets = letters[1:4])
+  )
+})
+
+test_that("map_batches with explicit schema", {
+  fun_with_dots <- function(batch, first_col, first_col_val) {
+    record_batch(
+      !! first_col := first_col_val,
+      b = batch$a$cast(float64())
+    )
+  }
+
+  empty_reader <- RecordBatchReader$create(
+    batches = list(),
+    schema = schema(a = int32())
+  )
+  expect_equal(
+    map_batches(
+      empty_reader,
+      fun_with_dots,
+      "first_col_name",
+      "first_col_value",
+      .schema = schema(first_col_name = string(), b = float64())
+    )$read_table(),
+    arrow_table(first_col_name = character(), b = double())
+  )
+
+  reader <- RecordBatchReader$create(
+    batches = list(
+      record_batch(a = 1, b = "two"),
+      record_batch(a = 2, b = "three")
+    )
+  )
+  expect_equal(
+    map_batches(
+      reader,
+      fun_with_dots,
+      "first_col_name",
+      "first_col_value",
+      .schema = schema(first_col_name = string(), b = float64())
+    )$read_table(),
+    arrow_table(
+      first_col_name = c("first_col_value", "first_col_value"),
+      b = as.numeric(1:2)
+    )
+  )
+})
+
+test_that("map_batches without explicit schema", {
+  fun_with_dots <- function(batch, first_col, first_col_val) {
+    record_batch(
+      !! first_col := first_col_val,
+      b = batch$a$cast(float64())
+    )
+  }
+
+  empty_reader <- RecordBatchReader$create(
+    batches = list(),
+    schema = schema(a = int32())
+  )
+  expect_error(
+    map_batches(
+      empty_reader,
+      fun_with_dots,
+      "first_col_name",
+      "first_col_value"
+    )$read_table(),
+    "Can't infer schema"
+  )
+
+  reader <- RecordBatchReader$create(
+    batches = list(
+      record_batch(a = 1, b = "two"),
+      record_batch(a = 2, b = "three")
+    )
+  )
+  expect_equal(
+    map_batches(
+      reader,
+      fun_with_dots,
+      "first_col_name",
+      "first_col_value"
+    )$read_table(),
+    arrow_table(
+      first_col_name = c("first_col_value", "first_col_value"),
+      b = as.numeric(1:2)
+    )
+  )
 })
 
 test_that("head/tail", {
@@ -758,24 +895,6 @@ test_that("Scanner$ScanBatches", {
   batches <- ds$NewScan()$Finish()$ScanBatches()
   table <- Table$create(!!!batches)
   expect_equal(as.data.frame(table), rbind(df1, df2))
-
-  expect_deprecated(ds$NewScan()$UseAsync(TRUE), paste(
-    "The function",
-    "'UseAsync' is deprecated and will be removed in a future release."
-  ))
-  expect_deprecated(ds$NewScan()$UseAsync(FALSE), paste(
-    "The function",
-    "'UseAsync' is deprecated and will be removed in a future release."
-  ))
-
-  expect_deprecated(Scanner$create(ds, use_async = TRUE), paste(
-    "The parameter 'use_async' is deprecated and will be removed in a future",
-    "release."
-  ))
-  expect_deprecated(Scanner$create(ds, use_async = FALSE), paste(
-    "The parameter 'use_async' is deprecated and will be removed in a future",
-    "release."
-  ))
 })
 
 test_that("Scanner$ToRecordBatchReader()", {
@@ -1000,3 +1119,196 @@ test_that("Filter parquet dataset with is.na ARROW-15312", {
     df %>% filter(is.na(z)) %>% collect()
   )
 })
+
+test_that("FileSystemFactoryOptions with DirectoryPartitioning", {
+  parent_path <- make_temp_dir()
+  ds_path <- file.path(parent_path, "a_subdir")
+  write_dataset(mtcars, ds_path, partitioning = "cyl", hive = FALSE)
+  expect_equal(dir(parent_path), "a_subdir")
+  expect_equal(dir(ds_path), c("4", "6", "8"))
+  ds <- open_dataset(
+    parent_path,
+    partitioning = "cyl",
+    factory_options = list(partition_base_dir = ds_path)
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  # Add an invalid file
+  file.create(file.path(ds_path, "ASDFnotdata.pq"))
+  ds <- open_dataset(
+    parent_path,
+    partitioning = "cyl",
+    factory_options = list(
+      partition_base_dir = ds_path,
+      # open_dataset() fails if we don't exclude invalid
+      exclude_invalid_files = TRUE
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  ds <- open_dataset(
+    parent_path,
+    partitioning = "cyl",
+    factory_options = list(
+      partition_base_dir = ds_path,
+      # We can also ignore by prefix
+      selector_ignore_prefixes = "ASDF"
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  # Add a _$folder$ file (which Hadoop may write)
+  file.create(file.path(ds_path, "cyl_$folder$"))
+  ds <- open_dataset(
+    parent_path,
+    partitioning = "cyl",
+    factory_options = list(
+      partition_base_dir = ds_path,
+      # we can't ignore suffixes but we can exclude invalid
+      exclude_invalid_files = TRUE
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  # Now with a list of files
+  ds <- open_dataset(
+    dir(ds_path, recursive = TRUE, full.names = TRUE),
+    factory_options = list(
+      exclude_invalid_files = TRUE
+    )
+  )
+  expect_equal(
+    ds %>%
+      summarize(sum(gear)) %>%
+      collect() %>%
+      as.data.frame(),
+    mtcars %>%
+      summarize(sum(gear))
+  )
+})
+
+test_that("FileSystemFactoryOptions with HivePartitioning", {
+  parent_path <- make_temp_dir()
+  ds_path <- file.path(parent_path, "a_subdir")
+  write_dataset(mtcars, ds_path, partitioning = "cyl")
+  expect_equal(dir(parent_path), "a_subdir")
+  expect_equal(dir(ds_path), c("cyl=4", "cyl=6", "cyl=8"))
+  ds <- open_dataset(parent_path)
+
+  # With Hive partitioning, partition_base_dir isn't needed
+  expect_setequal(names(ds), names(mtcars))
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  # Add an invalid file
+  file.create(file.path(ds_path, "ASDFnotdata.pq"))
+  ds <- open_dataset(
+    parent_path,
+    factory_options = list(
+      # open_dataset() fails if we don't exclude invalid
+      exclude_invalid_files = TRUE
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  ds <- open_dataset(
+    parent_path,
+    factory_options = list(
+      # We can also ignore by prefix
+      selector_ignore_prefixes = "ASDF"
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+
+  # Add a _$folder$ file (which Hadoop may write)
+  file.create(file.path(ds_path, "cyl_$folder$"))
+  ds <- open_dataset(
+    parent_path,
+    factory_options = list(
+      # we can't ignore suffixes but we can exclude invalid
+      exclude_invalid_files = TRUE
+    )
+  )
+  expect_equal(
+    ds %>%
+      arrange(cyl) %>%
+      pull(cyl),
+    sort(mtcars$cyl)
+  )
+})
+
+test_that("FileSystemFactoryOptions input validation", {
+  expect_error(
+    open_dataset(dataset_dir, factory_options = list(other = TRUE)),
+    'Invalid factory_options: "other"'
+  )
+  expect_error(
+    open_dataset(
+      dataset_dir,
+      partitioning = "part",
+      factory_options = list(partition_base_dir = 42)
+    ),
+    "factory_options$partition_base_dir is not a string",
+    fixed = TRUE
+  )
+  expect_error(
+    open_dataset(dataset_dir, factory_options = list(selector_ignore_prefixes = 42)),
+    "factory_options$selector_ignore_prefixes must be a character vector",
+    fixed = TRUE
+  )
+  expect_error(
+    open_dataset(dataset_dir, factory_options = list(exclude_invalid_files = 42)),
+    "factory_options$exclude_invalid_files must be TRUE/FALSE",
+    fixed = TRUE
+  )
+
+  expect_warning(
+    open_dataset(hive_dir, factory_options = list(partition_base_dir = hive_dir)),
+    "factory_options$partition_base_dir is not meaningful for Hive partitioning",
+    fixed = TRUE
+  )
+
+  files <- dir(dataset_dir, full.names = TRUE, recursive = TRUE)
+  expect_error(
+    open_dataset(files, factory_options = list(selector_ignore_prefixes = "__")),
+    paste(
+      "Invalid factory_options for creating a Dataset from a vector",
+      'of file paths: "selector_ignore_prefixes"'
+    ),
+    fixed = TRUE
+  )
+})
