@@ -75,6 +75,7 @@ struct SourceNode : ExecNode {
                         {"node.label", label()},
                         {"node.output_schema", output_schema()->ToString()},
                         {"node.detail", ToString()}});
+    END_SPAN_ON_FUTURE_COMPLETION(span_, finished_);
     {
       // If another exec node encountered an error during its StartProducing call
       // it might have already called StopProducing on all of its inputs (including this
@@ -84,6 +85,7 @@ struct SourceNode : ExecNode {
       if (stop_requested_) {
         return Status::OK();
       }
+      started_ = true;
     }
 
     CallbackOptions options;
@@ -96,66 +98,77 @@ struct SourceNode : ExecNode {
       options.executor = executor;
       options.should_schedule = ShouldSchedule::IfDifferentExecutor;
     }
-    finished_ = Loop([this, executor, options] {
-                  std::unique_lock<std::mutex> lock(mutex_);
-                  int total_batches = batch_count_++;
-                  if (stop_requested_) {
-                    return Future<ControlFlow<int>>::MakeFinished(Break(total_batches));
-                  }
-                  lock.unlock();
+    ARROW_ASSIGN_OR_RAISE(Future<> scan_task, plan_->BeginExternalTask());
+    if (!scan_task.is_valid()) {
+      finished_.MarkFinished();
+      // Plan has already been aborted, no need to start scanning
+      return Status::OK();
+    }
+    auto fut = Loop([this, options] {
+                 std::unique_lock<std::mutex> lock(mutex_);
+                 if (stop_requested_) {
+                   return Future<ControlFlow<int>>::MakeFinished(Break(batch_count_));
+                 }
+                 lock.unlock();
 
-                  return generator_().Then(
-                      [=](const util::optional<ExecBatch>& maybe_batch)
-                          -> Future<ControlFlow<int>> {
-                        std::unique_lock<std::mutex> lock(mutex_);
-                        if (IsIterationEnd(maybe_batch) || stop_requested_) {
-                          stop_requested_ = true;
-                          return Break(total_batches);
-                        }
-                        lock.unlock();
-                        ExecBatch batch = std::move(*maybe_batch);
-
-                        if (executor) {
-                          auto status = task_group_.AddTask(
-                              [this, executor, batch]() -> Result<Future<>> {
-                                return executor->Submit([=]() {
-                                  outputs_[0]->InputReceived(this, std::move(batch));
-                                  return Status::OK();
-                                });
-                              });
-                          if (!status.ok()) {
-                            outputs_[0]->ErrorReceived(this, std::move(status));
-                            return Break(total_batches);
-                          }
-                        } else {
-                          outputs_[0]->InputReceived(this, std::move(batch));
-                        }
-                        lock.lock();
-                        if (!backpressure_future_.is_finished()) {
-                          EVENT(span_, "Source paused due to backpressure");
-                          return backpressure_future_.Then(
-                              []() -> ControlFlow<int> { return Continue(); });
-                        }
-                        return Future<ControlFlow<int>>::MakeFinished(Continue());
-                      },
-                      [=](const Status& error) -> ControlFlow<int> {
-                        // NB: ErrorReceived is independent of InputFinished, but
-                        // ErrorReceived will usually prompt StopProducing which will
-                        // prompt InputFinished. ErrorReceived may still be called from a
-                        // node which was requested to stop (indeed, the request to stop
-                        // may prompt an error).
-                        std::unique_lock<std::mutex> lock(mutex_);
-                        stop_requested_ = true;
-                        lock.unlock();
-                        outputs_[0]->ErrorReceived(this, error);
-                        return Break(total_batches);
-                      },
-                      options);
-                }).Then([&](int total_batches) {
-      outputs_[0]->InputFinished(this, total_batches);
-      return task_group_.End();
-    });
-    END_SPAN_ON_FUTURE_COMPLETION(span_, finished_, this);
+                 return generator_().Then(
+                     [=](const util::optional<ExecBatch>& maybe_morsel)
+                         -> Future<ControlFlow<int>> {
+                       std::unique_lock<std::mutex> lock(mutex_);
+                       if (IsIterationEnd(maybe_morsel) || stop_requested_) {
+                         return Break(batch_count_);
+                       }
+                       lock.unlock();
+                       bool use_legacy_batching = plan_->UseLegacyBatching();
+                       ExecBatch morsel = std::move(*maybe_morsel);
+                       int64_t morsel_length = static_cast<int64_t>(morsel.length);
+                       if (use_legacy_batching || morsel_length == 0) {
+                         // For various reasons (e.g. ARROW-13982) we pass empty batches
+                         // through
+                         batch_count_++;
+                       } else {
+                         int num_batches = static_cast<int>(
+                             bit_util::CeilDiv(morsel_length, ExecPlan::kMaxBatchSize));
+                         batch_count_ += num_batches;
+                       }
+                       RETURN_NOT_OK(plan_->ScheduleTask([=]() {
+                         int64_t offset = 0;
+                         do {
+                           int64_t batch_size = std::min<int64_t>(
+                               morsel_length - offset, ExecPlan::kMaxBatchSize);
+                           // In order for the legacy batching model to work we must
+                           // not slice batches from the source
+                           if (use_legacy_batching) {
+                             batch_size = morsel_length;
+                           }
+                           ExecBatch batch = morsel.Slice(offset, batch_size);
+                           offset += batch_size;
+                           outputs_[0]->InputReceived(this, std::move(batch));
+                         } while (offset < morsel.length);
+                         return Status::OK();
+                       }));
+                       lock.lock();
+                       if (!backpressure_future_.is_finished()) {
+                         EVENT(span_, "Source paused due to backpressure");
+                         return backpressure_future_.Then(
+                             []() -> ControlFlow<int> { return Continue(); });
+                       }
+                       return Future<ControlFlow<int>>::MakeFinished(Continue());
+                     },
+                     [=](const Status& error) -> ControlFlow<int> {
+                       outputs_[0]->ErrorReceived(this, error);
+                       return Break(batch_count_);
+                     },
+                     options);
+               })
+                   .Then(
+                       [this, scan_task](int total_batches) mutable {
+                         outputs_[0]->InputFinished(this, total_batches);
+                         scan_task.MarkFinished();
+                         finished_.MarkFinished();
+                       },
+                       {}, options);
+    if (!executor && finished_.is_finished()) return finished_.status();
     return Status::OK();
   }
 
@@ -196,17 +209,18 @@ struct SourceNode : ExecNode {
   void StopProducing() override {
     std::unique_lock<std::mutex> lock(mutex_);
     stop_requested_ = true;
+    if (!started_) {
+      finished_.MarkFinished();
+    }
   }
 
-  Future<> finished() override { return finished_; }
-
  private:
   std::mutex mutex_;
   int32_t backpressure_counter_{0};
   Future<> backpressure_future_ = Future<>::MakeFinished();
   bool stop_requested_{false};
+  bool started_ = false;
   int batch_count_{0};
-  util::AsyncTaskGroup task_group_;
   AsyncGenerator<util::optional<ExecBatch>> generator_;
 };
 
@@ -231,7 +245,7 @@ struct TableSourceNode : public SourceNode {
   static arrow::Status ValidateTableSourceNodeInput(const std::shared_ptr<Table> table,
                                                     const int64_t batch_size) {
     if (table == nullptr) {
-      return Status::Invalid("TableSourceNode node requires table which is not null");
+      return Status::Invalid("TableSourceNode requires table which is not null");
     }
 
     if (batch_size <= 0) {
