@@ -114,7 +114,7 @@ void ConnectivityTest::TestBrokenConnection() {
 //------------------------------------------------------------
 // Tests of data plane methods
 
-void DataTest::SetUp() {
+void DataTest::SetUpTest() {
   server_ = ExampleTestServer();
 
   ASSERT_OK_AND_ASSIGN(auto location, Location::ForScheme(transport(), "127.0.0.1", 0));
@@ -123,7 +123,7 @@ void DataTest::SetUp() {
 
   ASSERT_OK(ConnectClient());
 }
-void DataTest::TearDown() {
+void DataTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
 }
@@ -237,6 +237,21 @@ void DataTest::TestDoGetLargeBatch() {
   Ticket ticket{"ticket-large-batch-1"};
   CheckDoGet(ticket, expected_batches);
 }
+// Ensure FlightDataStream/RecordBatchStream::Close errors are propagated
+void DataTest::TestFlightDataStreamError() {
+  Ticket ticket{"ticket-stream-error"};
+
+  ASSERT_OK_AND_ASSIGN(auto stream, client_->DoGet(ticket));
+  Status status;
+  while (true) {
+    FlightStreamChunk chunk;
+    status = stream->Next().Value(&chunk);
+    if (!chunk.data) break;
+    if (!status.ok()) break;
+  }
+  EXPECT_RAISES_WITH_MESSAGE_THAT(IOError, ::testing::HasSubstr("Expected error"),
+                                  status);
+}
 void DataTest::TestOverflowServerBatch() {
   // Regression test for ARROW-13253
   // N.B. this is rather a slow and memory-hungry test
@@ -637,14 +652,14 @@ class DoPutTestServer : public FlightServerBase {
   friend class DoPutTest;
 };
 
-void DoPutTest::SetUp() {
+void DoPutTest::SetUpTest() {
   ASSERT_OK_AND_ASSIGN(auto location, Location::ForScheme(transport(), "127.0.0.1", 0));
   ASSERT_OK(MakeServer<DoPutTestServer>(
       location, &server_, &client_,
       [](FlightServerOptions* options) { return Status::OK(); },
       [](FlightClientOptions* options) { return Status::OK(); }));
 }
-void DoPutTest::TearDown() {
+void DoPutTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
   reinterpret_cast<DoPutTestServer*>(server_.get())->batches_.clear();
@@ -865,14 +880,14 @@ Status AppMetadataTestServer::DoPut(const ServerCallContext& context,
   return Status::OK();
 }
 
-void AppMetadataTest::SetUp() {
+void AppMetadataTest::SetUpTest() {
   ASSERT_OK_AND_ASSIGN(auto location, Location::ForScheme(transport(), "127.0.0.1", 0));
   ASSERT_OK(MakeServer<AppMetadataTestServer>(
       location, &server_, &client_,
       [](FlightServerOptions* options) { return Status::OK(); },
       [](FlightClientOptions* options) { return Status::OK(); }));
 }
-void AppMetadataTest::TearDown() {
+void AppMetadataTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
 }
@@ -1044,14 +1059,14 @@ class IpcOptionsTestServer : public FlightServerBase {
   }
 };
 
-void IpcOptionsTest::SetUp() {
+void IpcOptionsTest::SetUpTest() {
   ASSERT_OK_AND_ASSIGN(auto location, Location::ForScheme(transport(), "127.0.0.1", 0));
   ASSERT_OK(MakeServer<IpcOptionsTestServer>(
       location, &server_, &client_,
       [](FlightServerOptions* options) { return Status::OK(); },
       [](FlightClientOptions* options) { return Status::OK(); }));
 }
-void IpcOptionsTest::TearDown() {
+void IpcOptionsTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
 }
@@ -1232,7 +1247,7 @@ class CudaDataTest::Impl {
   std::shared_ptr<cuda::CudaContext> context;
 };
 
-void CudaDataTest::SetUp() {
+void CudaDataTest::SetUpTest() {
   ASSERT_OK_AND_ASSIGN(auto manager, cuda::CudaDeviceManager::Instance());
   ASSERT_OK_AND_ASSIGN(auto device, manager->GetDevice(0));
   ASSERT_OK_AND_ASSIGN(auto context, device->GetContext());
@@ -1251,7 +1266,7 @@ void CudaDataTest::SetUp() {
       [](FlightClientOptions* options) { return Status::OK(); }, impl_->device,
       impl_->context));
 }
-void CudaDataTest::TearDown() {
+void CudaDataTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
 }
@@ -1353,8 +1368,8 @@ void CudaDataTest::TestDoExchange() {
 
 #else
 
-void CudaDataTest::SetUp() {}
-void CudaDataTest::TearDown() {}
+void CudaDataTest::SetUpTest() {}
+void CudaDataTest::TearDownTest() {}
 void CudaDataTest::TestDoGet() { GTEST_SKIP() << "Arrow was built without ARROW_CUDA"; }
 void CudaDataTest::TestDoPut() { GTEST_SKIP() << "Arrow was built without ARROW_CUDA"; }
 void CudaDataTest::TestDoExchange() {
@@ -1437,17 +1452,29 @@ class ErrorHandlingTestServer : public FlightServerBase {
     }
     return Status::NotImplemented("NYI");
   }
+
+  Status DoPut(const ServerCallContext& context,
+               std::unique_ptr<FlightMessageReader> reader,
+               std::unique_ptr<FlightMetadataWriter> writer) override {
+    return MakeFlightError(FlightStatusCode::Unauthorized, "Unauthorized", "extra info");
+  }
+
+  Status DoExchange(const ServerCallContext& context,
+                    std::unique_ptr<FlightMessageReader> reader,
+                    std::unique_ptr<FlightMessageWriter> writer) override {
+    return MakeFlightError(FlightStatusCode::Unauthorized, "Unauthorized", "extra info");
+  }
 };
 }  // namespace
 
-void ErrorHandlingTest::SetUp() {
+void ErrorHandlingTest::SetUpTest() {
   ASSERT_OK_AND_ASSIGN(auto location, Location::ForScheme(transport(), "127.0.0.1", 0));
   ASSERT_OK(MakeServer<ErrorHandlingTestServer>(
       location, &server_, &client_,
       [](FlightServerOptions* options) { return Status::OK(); },
       [](FlightClientOptions* options) { return Status::OK(); }));
 }
-void ErrorHandlingTest::TearDown() {
+void ErrorHandlingTest::TearDownTest() {
   ASSERT_OK(client_->Close());
   ASSERT_OK(server_->Shutdown());
 }
@@ -1487,5 +1514,68 @@ void ErrorHandlingTest::TestGetFlightInfo() {
   }
 }
 
+void CheckErrorDetail(const Status& status) {
+  auto detail = FlightStatusDetail::UnwrapStatus(status);
+  ASSERT_NE(detail, nullptr) << status.ToString();
+  ASSERT_EQ(detail->code(), FlightStatusCode::Unauthorized);
+  ASSERT_EQ(detail->extra_info(), "extra info");
+}
+
+void ErrorHandlingTest::TestDoPut() {
+  // ARROW-16592
+  auto schema = arrow::schema({field("int64", int64())});
+  auto descr = FlightDescriptor::Path({""});
+  FlightClient::DoPutResult stream;
+  auto status = client_->DoPut(descr, schema).Value(&stream);
+  if (!status.ok()) {
+    ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(status));
+    return;
+  }
+
+  std::thread reader_thread([&]() {
+    std::shared_ptr<Buffer> out;
+    while (true) {
+      if (!stream.reader->ReadMetadata(&out).ok()) {
+        return;
+      }
+    }
+  });
+
+  auto batch = RecordBatchFromJSON(schema, "[[0]]");
+  while (true) {
+    status = stream.writer->WriteRecordBatch(*batch);
+    if (!status.ok()) break;
+  }
+
+  ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(status));
+  ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(stream.writer->Close()));
+  reader_thread.join();
+}
+
+void ErrorHandlingTest::TestDoExchange() {
+  // ARROW-16592
+  FlightClient::DoExchangeResult stream;
+  auto status = client_->DoExchange(FlightDescriptor::Path({""})).Value(&stream);
+  if (!status.ok()) {
+    ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(status));
+    return;
+  }
+
+  std::thread reader_thread([&]() {
+    while (true) {
+      if (!stream.reader->Next().ok()) return;
+    }
+  });
+
+  while (true) {
+    status = stream.writer->WriteMetadata(Buffer::FromString("foo"));
+    if (!status.ok()) break;
+  }
+
+  ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(status));
+  ASSERT_NO_FATAL_FAILURE(CheckErrorDetail(stream.writer->Close()));
+  reader_thread.join();
+}
+
 }  // namespace flight
 }  // namespace arrow
