@@ -981,8 +981,7 @@ Status ScalarFromJSON(const std::shared_ptr<DataType>& type,
   RETURN_NOT_OK(converter->AppendValue(json_doc));
   RETURN_NOT_OK(converter->Finish(&array));
   DCHECK_EQ(array->length(), 1);
-  ARROW_ASSIGN_OR_RAISE(*out, array->GetScalar(0));
-  return Status::OK();
+  return array->GetScalar(0).Value(out);
 }
 
 Status DictScalarFromJSON(const std::shared_ptr<DataType>& type,
